local MAJOR = "StarLibDriverCharacter-1.0"
local MINOR = 5
assert(LibStub, MAJOR.." requires LibStub") 
local DriverCharacter = LibStub:NewLibrary(MAJOR, MINOR)
if not DriverCharacter then return end
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibCore = LibStub("StarLibCore-1.0", true)
assert(LibCore, MAJOR .. " requires StarLibCore-1.0")
local LCDText = LibStub("StarLibLCDText-1.0", true)
assert(LCDText, MAJOR .. " requires StarLibLCDText-1.0")
local LibFont = LibStub("StarLibFont-1.0", true)
assert(LibFont, MAJOR .. " requires StarLibFont-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

local options
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not DriverCharacter.__index then
	DriverCharacter.__index = DriverCharacter
end

-- @name StarDriverCharacter:New
-- @usage StarDriverCharacter:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor App
-- @param environment Execution environment
-- @param name A name for this LCD
-- @param config The configuration
-- @param errorLevel Error verbosity
-- @return A new StarDriverCharacter object

function DriverCharacter:New(visitor, environment, name, config, errorLevel)
	
	assert(type(name) == "string", MAJOR .. ": Invalid name")
	assert(type(config[name]) == "table", MAJOR .. ": " .. name .. ": Invalid config")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}		
	end

	setmetatable(obj, self)
	
	obj.environment = environment
	obj.core = LibCore:New(obj, obj.environment, name, config, "text", errorLevel)
	obj.error = LibError:New(MAJOR, errorLevel)
	--local rows, cols = config[name].rows, config[name].cols
	obj.rows, obj.cols, obj.yres, obj.xres, obj.layers = config[name].rows or 1, config[name].cols or 16, config[name].yres or 8, config[name].xres or 7, config[name].layers or 1
	obj.lcd = LCDText:New(obj.core, obj.rows, obj.cols, obj.yres, obj.xres, obj.layers, errorLevel, obj.Blit, obj)
	obj.core.lcd = obj.lcd -- You must provide a LibCore object with an LCD object.
	obj.pixel = config[name].pixel or 3
	
	self.background = config.background
	
	obj.core:CFGSetup()
	obj.core:BuildLayouts()

	
	
	local frame = CreateFrame("Frame")
	obj.frame = frame
	frame:SetFrameStrata("BACKGROUND")
	frame:SetFrameLevel(0)
	frame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground", 
		tile = true, 
		tileSize = obj.pixel,
		insets = {left = 0, right = 0, top = 0, bottom = 0},
	})
	frame:SetWidth(obj.lcd.DCOLS * obj.lcd.XRES * obj.pixel)
	frame:SetHeight(obj.lcd.DROWS * obj.lcd.YRES * obj.pixel)
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT", config[name].row, config[name].col or 0)
	StarTip:Print(config[name].row, config[name].col)
	StarTip:Print(config[name].row, "test")
	frame:SetAlpha(1)
	frame:Hide()
	
	obj.frames = LibBuffer:New(MAJOR .. " " .. name, obj.lcd.DROWS * obj.lcd.DCOLS * obj.lcd.YRES * obj.lcd.XRES, false, errorLevel)
	for i = 0, obj.lcd.DCOLS * obj.lcd.DROWS * obj.lcd.XRES * obj.lcd.YRES - 1 do
		local frame = CreateFrame("Frame")
		local row, col = LibUtils:GetCoords(i, obj.lcd.DCOLS * obj.lcd.XRES)
		obj.frames.buffer[i] = frame
		frame:SetParent(obj.frame)
		frame:SetFrameStrata("BACKGROUND")
		frame:SetFrameLevel(1)
		frame:SetBackdrop({
			bgFile = "Interface\\ChatFrame\\ChatFrameBackground", 
			tile = true, 
			tileSize = pixel,
			insets = {left = 0, right = 0, top = 0, bottom = 0},
		})
		frame:SetWidth(obj.pixel)
		frame:SetHeight(obj.pixel)
		frame:ClearAllPoints()
		frame:SetPoint("BOTTOMLEFT", col * obj.pixel, row * obj.pixel)
		frame:SetBackdropColor(random(2) - 1, random(2) - 1, random(2) - 1)
		frame:Show()
	end
	
	obj.buffer = LibBuffer:New(MAJOR .. " " .. name .. " buffer", obj.lcd.DCOLS * obj.lcd.XRES * obj.lcd.DROWS * obj.lcd.YRES, 0, errorLevel)
	
	obj.timer = LibTimer:New(MAJOR .. " " .. name, 100, true, obj.Update, obj, errorLevel)
	obj.timer.timer = 0xdead
	obj:Clear()
	
	return obj
end

-- Delete an object
-- @name StarDriverCharacter:Del
-- @usage StarDriverCharacter:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function DriverCharacter:Del(lcd)
	if not lcd then
		lcd = self
	end
	pool[lcd] = true
	lcd.core:Del()
	lcd.error:Del()
	lcd.timer:Del()
	lcd:Hide()
end

-- Show the tooltip
-- @name StarDriverCharacter:Show
-- @usage StarDriverCharacter:Show()
-- @return Nothing
function DriverCharacter:Show()
	self.frame:Show()
	self.core:Start()
	self:Move("CENTER", 0, 0)
	self.timer:Start()
end

-- Hide the display
-- @name :Hide
-- @usage :Hide()
-- @return Nothing
function DriverCharacter:Hide()
	self.core:Stop()
	self.frame:Hide()
	self.timer:Stop()
end

function DriverCharacter:Clear()
	self.lcd:Clear()
	self.lcd.DisplayFB:Fill("#")
	self.lcd:TextBlit(0, 0, self.lcd.DCOLS, self.lcd.DROWS)
end

-- Move the tooltip
-- @name StarDriverCharacter:Move
-- @usage StarDriverCharacter(anchor, x, y)
-- @param anchor Frame anchor, i.e: CENTER.
-- @param x X position
-- @param y Y position
-- @return Nothing
function DriverCharacter:Move(anchor, x, y)
	self.frame:ClearAllPoints()
	self.frame:SetPoint(anchor, x, y)
end

function DriverCharacter:Blit(obj, r, c, buffer, len)
	for i = c, c + len - 1 do
		obj:SetCell(r, i, (buffer[i - c] or ' '):sub(1, 1))
	end
end

function DriverCharacter:RebuildOpts(visitor, db)
	return LibCore:RebuildOpts(visitor, db)
end

function DriverCharacter:KeyEvent(modifier, up)
	self.core:KeyEvent(modifier, up)
end

function DriverCharacter:SetCell(row, col, char, bold)
	local chr
	if(bold) then
		chr = LibFont.Font_6x8_bold[strbyte(char) + 1];
	else
		chr = LibFont.Font_6x8[strbyte(char) + 1];
	end
	
	for i = 1, #chr do
		local r, c = LibUtils:GetCoords(i, self.lcd.XRES)
	end
	for y = 0 , self.lcd.YRES - 1 do
		local mask = bit.lshift(1, self.lcd.XRES)
		for x = 0, self.lcd.XRES - 1 do
			mask = bit.rshift(mask, 1)
			if bit.band(chr[self.lcd.YRES - y], mask) == 0 then
				self.buffer.buffer[(row * self.lcd.YRES + y) * self.lcd.LCOLS * self.lcd.XRES + col * self.lcd.XRES + x] = 1
			else
				self.buffer.buffer[(row * self.lcd.YRES + y) * self.lcd.LCOLS * self.lcd.XRES + col * self.lcd.XRES + x] = 0
			end
		end
	end
end

function DriverCharacter:Update()
	for i = 0, self.lcd.DROWS * self.lcd.DCOLS * self.lcd.YRES * self.lcd.XRES - 1 do
		if self.buffer.buffer[i] == 0 then
			self.frames.buffer[i]:SetBackdropColor(1, 1, 1)
		else
			self.frames.buffer[i]:SetBackdropColor(0, 0, 0)
		end
	end
end